#include <algorithm>
#include <fstream>
#include <istream>
#include <iostream>
#include <limits>
#include <list>
#include <memory>
#include <stdlib.h>
#include <string>
#include <utility>
#include <vector>
#include <sdsl/construct.hpp>

#include "../../index/mti_lzss.hpp"

using namespace std;

void printUsage(const char *progname) {
	cerr << "Usage: " << progname << " MAX_REPLEN MAX_OCC OCC_DENS [Files]+" << endl;
	cerr << "Program finds repeats in concatenated file contents. " << endl;
	cerr << "Repeats will be printed to screen, separated by a newline." << endl;
	cerr << "PARAMETERS:" << endl;
	cerr << "- MAX_REPLEN: Maximal string length of repeats (inclusive)" << endl;
	cerr << "- REPLEN_DENS: Repeat Length Density: Instead that for all Repeat Lengths " << endl;
	cerr << "               a repeats get generated, only for MAX_REPLEN / REPLEN_DENS " << endl;
	cerr << "               repeat lengths (with different occurence counts) strings " << endl;
	cerr << "               will be printed." << endl;
	cerr << "- MAX_OCC: Maximal Occurence count of the repeats (inclusive)," << endl;
	cerr << "           each Repeats that occur more often are discarded" << endl;
	cerr << "- OCC_DENS: Occurence Density: For each repeat len and occurence " << endl;
	cerr << "            count, at most 1 repeat gets printed. To avoid printing " << endl;
	cerr << "            a repeat for each occurence count of a certain repeat len" << endl;
	cerr << "            One uses density to merge OCC_DENS directly following " << endl;
	cerr << "            occurence counts to 1, so at most MAX_OCC / OCC_DENS " << endl;
	cerr << "            repeats are printed for that repeat len. Default 1." << endl;
} 

int main(int argc, char **argv) {
	typedef typename mti_lzss::size_type size_type;
	//fetch maximal repeat length, maximal occurences and occurence density
	size_type max_replen, replen_dens, max_occ, occ_dens;
	size_type *inputs[] = {&max_replen, &replen_dens, &max_occ, &occ_dens};
	if (argc < 5) {
		printUsage(argv[0]);
		return 1;
	}
	int arg = 1;
	while (arg < 5) {
		int input = atoi(argv[arg]);
		if (input <= 0) {
			printUsage(argv[0]);
			return 1;
		}
		*(inputs[arg++ - 1]) = (size_type)input;
	}
	unsigned char *textptr = NULL;
	vector<size_type> offsets;
	{
		//create a list containing streams to files to be concatenated
		list<unique_ptr<istream>> streamlist;
		while (arg < argc) {
			//open a ifstream for each file
			ifstream *s = new ifstream(argv[arg]);
			if (!s->good()) {
				printUsage(argv[0]);
				cerr << "Unable to open file " << argv[arg] << endl;
				return 1;
			}
			streamlist.push_back(unique_ptr<istream>( s ));
			arg++;			
		}
		//concatenate all texts to one big text
		textptr = mti_lzss::concatTextsWithSentinels(
		        streamlist.begin(), streamlist.end(), offsets );
		if (textptr == NULL) {
			printUsage(argv[0]);
			cerr << ((streamlist.size() > 0) 
			        ? "Unable to concatenate texts"
			        : "Please pass at least 1 text file") << endl;
			return 1;
		}
	}
	//by now, concatenated text and offsets in text which separate text parts
	// are created, so create SA and LCP
	//put text to cache
	string tmp_file = sdsl::ram_file_name(sdsl::util::to_string(
		sdsl::util::pid())+"_"+sdsl::util::to_string(sdsl::util::id()));
	sdsl::store_to_file((const char *)textptr, tmp_file);
	delete[] textptr;
	sdsl::cache_config config;
	config.dir = "@";

	//transform text to a int_vector in cache
	sdsl::int_vector<8> text;
	sdsl::load_vector_from_file(text, tmp_file, sizeof(unsigned char));
	sdsl::store_to_cache(text, sdsl::conf::KEY_TEXT, config);
	sdsl::register_cache_file(sdsl::conf::KEY_TEXT, config);

	//finally, calculate SA and LCP
	sdsl::construct_sa<8>(config);
	sdsl::construct_lcp_PHI<8>(config);

	//and fetch both out
	sdsl::int_vector_buffer<> sa (sdsl::cache_file_name(sdsl::conf::KEY_SA, config));
	sdsl::int_vector_buffer<> lcp(sdsl::cache_file_name(sdsl::conf::KEY_LCP, config));

	//After all caching stuff, finally do substring analysis
	vector<size_type> repeat_off(max_replen+1, 0); //offset in text where repeat starts
	vector<size_type> repeat_occ(max_replen+1, 0); //number of occurences of a repeat
	//create a matrix indicating if a repeat of a certain length and a number of 
	//occurences already was printed
	size_type strings_per_occcount = max_replen / replen_dens + 1;
	size_type strings_per_replen   = max_occ / occ_dens + 1;
	vector<vector<bool>> string_used(strings_per_occcount, vector<bool>(strings_per_replen, false));

	size_type occurences = 0; //occurences of actual text without first text occurence
	size_type last_replen = 0;
	size_type i = sa.size() + 1;
	while (i--) {
		occurences++;
		size_type replen = min((size_type)lcp[i], max_replen);
		if (replen > last_replen) {
			repeat_occ[last_replen] = occurences;
			do {
				++last_replen;
				repeat_off[last_replen] = sa[i];
				repeat_occ[last_replen] = 0;
			} while (last_replen < replen);
			occurences = 0;
		}
		else if (replen < last_replen) {
			do {
				//check if string is already used as representant
				size_type rpl_thin = last_replen / replen_dens;
				size_type occ_thin = (occurences + 1) / occ_dens;
				if (occurences < max_occ && !string_used[rpl_thin][occ_thin]) {
					//output string
					size_type str_off = repeat_off[last_replen];
					for (size_type j = 0; j < last_replen; j++) {
						cout << (char)(text[str_off + j]);
					}
					cout << '\0';
					//and mark that representant already is used
					string_used[rpl_thin][occ_thin] = true;
				}
				occurences += repeat_occ[--last_replen];
			} while (last_replen > replen);
		}
	}
	cout << '\0';
	return 0;
}
