#ifndef __SI_LZSS_H
#define __SI_LZSS_H

#include <list>
#include <stdint.h>
#include <string>
#include <sdsl/io.hpp>
#include <sdsl/vectors.hpp>
#include <sdsl/wavelet_trees.hpp>

#include "lzss_coding.hpp"
#include "patricia_trie.hpp"

//! Self - index based on Lempel - Ziv- Storer - Szymanski Coding
/*!	
 *	This self index mainly was inspired by preprint 
 *	,,A Faster Grammar - Based Self - Index'', namely the 
 *	rough description about Kreft and Navarro's self index.
 *
 *	Main difference between this self - index and that described
 *	by Kreft and Navarro is that this one searches secondary 
 *	occurences on a different way, covering same space and time 
 *	bounds as that by Kreft and Navarro, but using 2 arrays and a
 *	wavelet tree instead.
 *
 * \par Space complexity
 *      \f$ 4 n' \log n + \Order{n' \log n'} \f$
 *      where n is length of text and n' is count of lzss - phrases.
 *
 *	\par References
 *      [1] T. Gagie, P. Gawrychowski, J. Kärkkäinen, Y. Nekrich, S.J. Puglisi:
 *          ,,A Faster Grammar - Based Self-Index'',
 *          Preprint Sebtember 28, 2012.
 *	[2] S. Kreft, G. Navarro:
 *	    ,,On Compressing and Indexing Repetitive Sequences'',
 *	    Department of Computer Science, University of Chile
 */
class csi_lzss {
	public:
		typedef sdsl::int_vector<>::size_type	size_type;
	private:
		//lempel - ziv coding and utility
		lzss_coding coding;
		sdsl::int_vector<> p_src_p; //phrase source phrase:
			//a link to the phrase which contains owns phrase source.

		patricia_trie pt_phrasesuffixes; //on trie for the suffixes of phrases
		patricia_trie pt_revphrases; //and another one for the reversed phrases

		sdsl::wt_int<> rs_grid; //a grid to associate reversed phrases with suffixes

		//handling secondary occurences when finding a pattern
		sdsl::int_vector<> p_ord_src;      //phrases ordered by source(p)
		sdsl::int_vector<> p_ord_srcend;   //phrases ordered by source(p) + length(p)
		sdsl::wt_int<>     se_grid;  //a grid what associates source - ordered phrases with
					     //their (encoded) source + length - values
			

		// Construction method
		void construct(const unsigned char *text, const size_type t_len);

		//! a function to extract characters from LZSS coding
		/*!
		 *	\param p         the phrase at which the extraction should start.
		 *	\param skip      how much letters from the beginning of the phrase 
		 *	                 should be skipped.
		 *	\param count     count of letters to extract.
		 *	\param extractOn a callback function what receives the next extracted
		 *	                 character and should return, if extraction can stop
		 *	                 or should go on.
		 *	\returns         true if all count characters were extracted
		 *	                 (meaning that none call of extractOn returned false),
		 *	                 or false if extraction stopped prematurely.
		 */
		bool extract(size_type p, size_type skip, size_type count, 
		             std::function<bool (unsigned char)> extractOn) const;

	public:
		//! Empty Constructor
		csi_lzss() {}

		//! Constructs this self - index.
		/*!
		 *	\param text  text from which to create the index.
		 *	             text[t_len] = 0 and for i<len text[i] != 0.
		 *	\param t_len length of text.
		 *	\par Time Complexity
		 *	     \f$ \Order{n + n' \log n'} \f$
		 */
		csi_lzss(const unsigned char *text, const size_type t_len) {
			construct(text, t_len);
		}

		//! Constructs this self - index.
		/*!
		 *	\param text  text from which to create the index.
		 *	\par Time Complexity
		 *	     \f$ \Order{n + n' \log n'} \f$
		 */
		csi_lzss(const std::string &text) {
			construct((const unsigned char *)text.data(), text.length());
		}

		//! Returns the length of the text covered by internal LZSS coding.
		size_type textlen() const {
			return (coding.empty()) ? 0 : *(--coding.end());
		}

		//! Extracts text[i .. i + m - 1] and saves it to buffer.
		/*!
		 *	\param buff buffer being able to store m characters.
		 *	\param i    start index from which to extract characters.
		 *	\param m    length of text to be extracted.
		 *	\par Time Complexity
		 *	     \f$ \Order{m l} \f$ where l is height of lzss parsing.
		 */
		void extract(unsigned char *buff, size_type i, size_type m) const;

		//! Extracts text[i .. i + m - 1].
		/*!
		 *	\param i    start index from which to extract characters.
		 *	\param m    length of text to be extracted.
		 *	\returns    the extracted text.
		 *	\par Time Complexity
		 *	     \f$ \Order{m l} \f$ where l is height of lzss parsing.
		 */
		std::string extract(size_type i, size_type m) const {
			std::vector<unsigned char> buff(m); //create a vector buffer
			extract(buff.data(), i, m);         //extract
			return std::string((const char *)buff.data(), m); //and make string out of it
		}

		//! Searches for positions where pattern is substring of text.
		/*!
		 *	\param p   pattern to search for.
		 *	\param m   length of pattern.
		 *	\returns   a list with positions in text where pattern was found.
		 *	\par Time Complexity
		 *	     \f$ \Order{m^2 l + occ \log n'} \f$ 
		 *	     where l is height of lzss parsing and occ is occurence count.
		 */
		std::list<size_type> findPattern(const unsigned char *p, size_type m);

		//! Searches for positions where pattern is substring of text.
		/*!
		 *	\param p   pattern to search for.
		 *	\returns   a list with positions in text where pattern was found.
		 *	\par Time Complexity
		 *	     \f$ \Order{m^2 l + occ \log n'} \f$ 
		 *	     where l is height of lzss parsing and occ is occurence count.
		 */
		std::list<size_type> findPattern(const std::string &p) {
			return findPattern((const unsigned char *)p.data(), p.length());
		}

        	//! Serializes the data structure into the given ostream
        	size_type serialize(std::ostream& out, sdsl::structure_tree_node* v=nullptr,
                            std::string name="") const;

		//! Loads the data structure from the given istream.
		void load(std::istream& in);
};

#endif
