#include "lzss_coding.hpp"
#include <algorithm>
#include <assert.h>
#include <limits>
#include <sdsl/bits.hpp>
#include <sdsl/util.hpp>

typedef typename lzss_coding::size_type size_type;

lzss_coding::lzss_coding() {}

//// Construction /////////////////////////////////////////////////////////////

size_type lzss_coding::parse_phrase(const unsigned char *text, 
		const size_type t_len, size_type pos,
		size_type psv_text, size_type nsv_text,
		size_type phrase) {
	
	size_type psv_len = 0, nsv_len = 0;
	if (psv_text != t_len) { //ensure a previous smaller text exists
		while (text[pos + psv_len] == text[psv_text + psv_len])	psv_len++;
	}
	if (nsv_text != t_len) { //ensure a next smaller text exists
		while (text[pos + nsv_len] == text[nsv_text + nsv_len]) nsv_len++;
	}

	//check if phrase is a value or not
	if (std::max(psv_len, nsv_len) <= 1) { //phrase contains a value
		src[phrase] = (size_type) text[pos];
		pos++;
	} 
	else if (psv_len > nsv_len) { //use psv_text as source
		src[phrase] = psv_text;
		pos += psv_len;
	}
	else { //use nsv_text as source
		src[phrase] = nsv_text;
		pos += nsv_len;
	}

	//save end position and let main algorithm do it's job
	pend[phrase] = pos;
	return pos;
}

lzss_coding::lzss_coding(const unsigned char *text, const size_type t_len, 
		const sdsl::int_vector<> &sa) {
	if (t_len == 0)
		return;

	//prepare psv_text and nsv_text (source and pend can be used as "temporary"
	// integer vectors)
	sdsl::int_vector<> &psv_text = src;
	sdsl::int_vector<> &nsv_text = pend;

	//create both vectors with enough space
	//psv_text[i] = t_len indicates that no previous smaller text exists,
	//psv_text[i] = t_len indicates that no next smaller text exists

	//ensure source array (= psv_text) has enough width to save characters
	psv_text = sdsl::int_vector<>(t_len, t_len, 
			std::max(sdsl::bits::hi(t_len) + 1, (uint32_t)std::numeric_limits<unsigned char>::digits));
	nsv_text = sdsl::int_vector<>(t_len, t_len, sdsl::bits::hi(t_len) + 1);

	//calculate psv_text and nsv_text
	for (size_type i = 1; i < sa.size(); i++) {
		size_type j = sa[i-1];
		size_type k = sa[i];
		while (k < j) {
			nsv_text[j] = k;
			j = psv_text[j];
			if (j == t_len) //no further previous smaller text exists
				break;
		}
		psv_text[k] = j;
	}

	//compute phrases. Since phrase <= i for all i, there's no problem reading and
	//writing from same vectors (psv_text = src; nsv_text = pend)
	size_type phrase = 0;
	for (size_type i = 0; 
	     i < t_len;
	     i = parse_phrase(text, t_len, i, psv_text[i], nsv_text[i], phrase++));

	//resize vectors, because they don't need all space
	src.resize(phrase);
	pend.resize(phrase);
}

//// SERIALIZATION AND LOAD ///////////////////////////////////////////////////

size_type lzss_coding::serialize(std::ostream& out, sdsl::structure_tree_node* v,
		std::string name) const {
	sdsl::structure_tree_node *child = sdsl::structure_tree::add_child(
		                                     v, name, sdsl::util::class_name(*this));
	size_type written_bytes = 0;
	written_bytes += src.serialize(out, child, "source");
	written_bytes += pend.serialize(out, child, "phraseend");
	sdsl::structure_tree::add_size(child, written_bytes);
	return written_bytes;
}

void lzss_coding::load(std::istream& in) {
	src.load(in);
	pend.load(in);
}
