#ifndef MTI_LZSS
#define MTI_LZSS

#include <algorithm>
#include <istream>
#include <ios>
#include <iterator>
#include <list>
#include <limits>
#include <memory>
#include <utility>
#include <vector>
#include <stdexcept>
#include <sdsl/io.hpp>

#include "csi_lzss.hpp"

//! A Index for multiple texts based on Compressed LZSS Self - Index.
/*!	
 *	This self index is based on the compressed lempel - ziv self - index 
 *      within same folder. It offers a method for searching a particular pattern
 *      in multiple strings.
 */

class mti_lzss {
	public:
		typedef csi_lzss::size_type	                   size_type;
		typedef std::list<std::pair<size_type, size_type>> occ_list; //list of occurences
		
		//! Concatenates the given text streams to one big text separated by unique sentinels.
		/*!
		 *	Texts mustn't contain '\0' - character, or function fails.
		 *	Also, there must be as much unused characters as texts,
		 *	or the function fails.
		 *
		 *	\param begin   iterator pointing on the first text stream to be used 
		 *	               in this index.
		 *	\param end     iterator pointing to the first element behind the 
		 *	               last text stream of this index.
		 *	\param offsets offsets where texts start. Last offset contains 
		 *	               length of whole concatenated text (including termination 
		 *	               character).
		 *	\returns       the concatenated text as a c - string, terminated by
		 *	               a 0 (not included in text length) or null on any problems.
		 *	               Do not forget to free space of text after work, using delete[].
		 */
		template<template<class> class It>
		static unsigned char *concatTextsWithSentinels( It<std::unique_ptr<std::istream>> begin,
		                                                It<std::unique_ptr<std::istream>> end, 
			                                        std::vector<size_type> &offsets);

	private:
		std::vector<size_type> offsets; //offsets when texts begin
		csi_lzss csi;                   //self - index used for pattern search
	public:
		//! Default Constructor
		mti_lzss() {}

		//! Constructs this self - index.
		/*!
		 *	\param begin  iterator pointing on the first text to be used 
		 *	              in this index.
		 *	\param end    iterator pointing to the first element behind the 
		 *	              last text of this index.
		 */
		template<template<class> class It>
		mti_lzss(It<std::unique_ptr<std::istream>> begin,
		         It<std::unique_ptr<std::istream>> end);

		//! Searches for positions where pattern is substring of text.
		/*!
		 *	\param p   pattern to search for.
		 *	\param m   length of pattern.
		 *	\returns   a list of pairs, where each pair contains
		 *	           (1) first, the text identifier of found occurence
		 *	           (2) second, the position of pattern in text
		 */
		occ_list findPattern(const unsigned char *p, size_type m);		

		//! Searches for positions where pattern is substring of text.
		/*!
		 *	\param p   pattern to search for.
		 *	\returns   a list of pairs, where each pair contains
		 *	           (1) first, the text identifier of found occurence
		 *	           (2) second, the position of pattern in text
		 */
		occ_list findPattern(const std::string &p) {
			return findPattern((const unsigned char *)p.data(), p.length());
		}

		//TODO: Add some text extract functionality in future

        	//! Serializes the data structure into the given ostream
        	size_type serialize(std::ostream& out, sdsl::structure_tree_node* v=nullptr,
                            std::string name="") const;

		//! Loads the data structure from the given istream.
		void load(std::istream& in);
};

//// CONCATENATION OF TEXTS ///////////////////////////////////////////////////

template<template<class> class It>
unsigned char *mti_lzss::concatTextsWithSentinels( It<std::unique_ptr<std::istream>> begin,
		                                   It<std::unique_ptr<std::istream>> end, 
			                           std::vector<mti_lzss::size_type> &offsets) {
	if (begin == end)	return NULL;

	//prepare offset array
	size_type num_texts = std::distance(begin, end);
	offsets.resize(num_texts+1);

	//iterate over streams and compute offsets
	offsets[0] = 0;
	size_type i = 1;
	size_type actual_offset = 0;
	for (auto it = begin; it != end; ++it) {
		auto &is = *it;
		if (!is->good()) {
			throw std::runtime_error("Unable to get stream size");
			return NULL;
		}

		is->seekg(0, std::ios_base::end);
		actual_offset += (size_type)is->tellg() + 1; //for sentinel character between strings
		is->seekg(0, std::ios_base::beg);

		offsets[i++] = actual_offset;
	}

	//allocate text
	unsigned char *text = new unsigned char[actual_offset];

	//iterate over streams again, and read everything into text
	i = 0;
	for (auto it = begin; it != end; ++it, ++i) {
		auto &is = *it;
		is->read((char *)(text + offsets[i]),  offsets[i+1] - offsets[i] - 1);
		if (!is->good()) {
			delete[] text;
			throw std::runtime_error("Unable to read stream");
			return NULL;
		}
	}

	//finally, try to find different sentinel characters not used in texts.
	//the '\0' is reserved for usage behind last text.
	std::vector<bool> char_used( std::numeric_limits<unsigned char>::max() + 1, false);
	for (size_type i = 0; i < num_texts; i++) {
		size_type start = offsets[i];
		size_type end   = offsets[i+1] - 1;
		while (start < end) {
			char_used[ text[ start++ ] ] = true;
		}
	}
	//check if '\0' already gets used in text
	if (char_used[0]) {
		delete[] text;
		return NULL;
	}

	//and insert remaining sentinel characters
	unsigned char sentinel = 0;
	for (size_type i = 1; i < num_texts; i++) {
		do {
			if (++sentinel >= char_used.size()) {
				delete[] text; //not enough sentinels available
				return NULL;
			}
		} while (char_used[sentinel]);

		text[ offsets[i] - 1] = sentinel;
	}
	text[ offsets[num_texts] - 1 ] = 0;
	return text;
}

//// CONSTRUCTION /////////////////////////////////////////////////////////////

template<template<class> class It>
mti_lzss::mti_lzss(It<std::unique_ptr<std::istream>> begin,
		   It<std::unique_ptr<std::istream>> end) {

	if (begin == end)	return;

	//build a string concatenation of all strings
	unsigned char *text = concatTextsWithSentinels( begin, end, offsets );
	if (text == NULL) {
		throw std::invalid_argument(
			"Texts mustn't contain \\0 character and must have as \
			much unused characters as texts are given." );
	}

	//build self - index
	csi = csi_lzss(text, offsets.back() - 1);
	//and free resources
	delete[] text;
}

#endif
