#ifndef __SI_PATRICIA_TRIE_H
#define __SI_PATRICIA_TRIE_H

#include <assert.h>
#include <functional>
#include <vector>
#include <map>
#include <stack>
#include <stdint.h>
#include <utility>
#include <sdsl/int_vector.hpp>
#include <sdsl/io.hpp>

//! A Patricia Trie that is able to save multiple strings that are equal.
/*! 	\author Uwe Baier
 *
 *	This is an implementation of a Patricia Trie that supports the 
 *	insertion of any strings (which don't necessarily need to be different).
 *	For the fact that multiple equal strings can be inserted, the 
 *	patricia trie is immutable after construction.
 *	One may note that strings mustn't contain '0' - character.
 *
 * 	\par Space complexity
 *      	\f$ \Order{n \log max(n, maxstrlen, \sigma)} \f$
 *      	where n is count of strings inserted and maxstrlen is length of longest string.
 *
 *	\par References
 *	[1] D. Morrison:
 *	    ,,PATRICIA-Practical algorithm to retrieve information coded in alphanumeric'',
 *	    Journal of the ACM 15 (1968) 514-534.
 */
class patricia_trie {
	public:
		typedef typename sdsl::int_vector<>::value_type  	value_type;
		typedef typename sdsl::int_vector<>::size_type		size_type;
	private:
		sdsl::int_vector<> lm_leaf; //index of leftmost leaf from this node
		sdsl::int_vector<> rm_leaf; //index of rightmost leaf from this node
		//TODO: remove lm leaf or rm leaf, since one array could be computed by the other
		// using left neighbor node (lm_leaf(n) = rm_leaf(left neighbor n) + 1)
		
		sdsl::int_vector<> e_lbl_len; //length of edge label
		//TODO: trim width of children to log(n)
		std::vector<std::map<unsigned char,size_type>> children; //map to indices of children

		sdsl::int_vector<> leafs;

		//utility for creation
		size_type createNode(size_type _lm_leaf, size_type _rm_leaf, 
		                     size_type _e_lbl_len, size_type &nodecount);

		void update_lm_leaf(std::stack<size_type> &&nodestack, 
		                    size_type old_lm_leaf_val, size_type new_lm_leaf_val);

		void update_rm_leaf(std::stack<size_type> &&nodestack,
		                    size_type old_rm_leaf_val, size_type new_rm_leaf_val);

		void insert(value_type i, 
		            std::function<unsigned char (value_type,size_type)> strChar,
		            std::function<size_type (value_type)> strLen,
		            sdsl::int_vector<> &nextleaf, size_type &nodecount);

		void sortLeafsLex(size_type n, size_type nodetextlen, size_type &l,
		                  std::function<size_type (value_type)> strLen,
		                  const sdsl::int_vector<> &nextleaf);

		void createEmptyTrie();
	public:
		//! Default constructor
		patricia_trie();

		//! Constructor for patricia trie. Strings mustn't contain '0' - character.
		/*! \param n          Number of strings to be inserted in the trie.
		    \param strChar    A function returning the j-th character of i-th string
		    \param strLen     A function returning the length of the i-th string
		    \param maxstrlen  Length of longest string
		    \par Time complexity
		         \f$ \Order{\sum_{i=0}^{n-1}strLen(i)} \f$
		*/
		patricia_trie(size_type n,
		              std::function<unsigned char (value_type i,size_type j)> strChar,
		              std::function<size_type (value_type i)> strLen,
		              size_type maxstrlen = 0);

		//! Function returns the lexicographic range of a prefix in this trie.
		/*! \param strChar      A function returning the j-th character of i-th string
		                        from construction.
		    \param charAtPrefix A function returning the j-th character of prefix.
		    \param prefixlen    Length of the prefix.
		    \returns            The lexicographic range of prefix on trie. Both bounds
		                        are inclusive. In case prefix is not contained in trie,
		                        function returns an illegal interval (e.g. [1,0]).
		    \par Time complexity
		         \f$ \Order{prefixlen} \f$
		*/
		std::pair<size_type,size_type> prefix_lex_range(
		               std::function<unsigned char (value_type i,size_type j)> strChar,
		               std::function<unsigned char (size_type j)> charAtPrefix, 
		               size_type prefixlen) const;

		//! Function returns the partial lexicographic range of a prefix in this trie.
		/*! \param charAtPrefix A function returning the j-th character of prefix.
		    \param prefixlen    Length of the prefix.
		    \returns            The partial lexicographic range of prefix on trie. 
		                        Since a patricia trie does not save all string characters
		                        on construction,the partial range will always be greater
		                        or equal to the real lexicographic range. Both bounds
		                        are inclusive. In case partial prefix is not contained 
		                        in trie, function returns an illegal interval (e.g. [1,0]).
		    \par Time complexity
		         \f$ \Order{prefixlen} \f$
		*/
		std::pair<size_type,size_type> partial_prefix_lex_range(
		               std::function<unsigned char (size_type j)> charAtPrefix, 
		               size_type prefixlen) const;

		//! Returns the index of the string with lexicographic order l.
		value_type lex2str(size_type l) const {
			assert(l < leafs.size());
			return leafs[l];
		}

		//! Returns the number of strings contained in this trie.
		size_type size() const {
			return leafs.size();
		}

        	//! Serializes the data structure into the given ostream
        	size_type serialize(std::ostream& out, sdsl::structure_tree_node* v=nullptr,
                            std::string name="") const;

		//! Loads the data structure from the given istream.
		void load(std::istream& in);
};

#endif
